# -----------------------------------------------------------------------------
# Copyright (c) 2021-2022 Science and Technology Facilities Council.
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""Module containing tests for the SymbolTables container functionality
of fparser2."""

import pytest
from fparser.two import Fortran2003
from fparser.two.symbol_table import SymbolTables, SymbolTable, SymbolTableError


def test_construction_addition_removal():
    """Check that we can create a SymbolTables instance, add a table to it,
    remove a table from it and query it."""
    tables = SymbolTables()
    assert tables._current_scope is None
    assert tables._symbol_tables == {}
    assert tables._enable_checks is False
    with pytest.raises(KeyError) as err:
        tables.lookup("missing")
    assert "missing" in str(err.value)
    # Add a symbol table
    table1 = tables.add("table1")
    assert isinstance(table1, SymbolTable)
    assert tables.lookup("taBLe1") is table1
    # We should not be able to add another table with the same name
    with pytest.raises(SymbolTableError) as err:
        tables.add("table1")
    assert (
        "table of top-level (un-nested) symbol tables already contains "
        "an entry for 'table1'" in str(err.value)
    )
    # Add a second table and then remove it
    table2 = tables.add("taBLe2")
    # Check that validation checks are disabled by default
    assert table2._checking_enabled is False
    assert tables.lookup("table2") is table2
    tables.remove("table2")
    with pytest.raises(KeyError) as err:
        tables.lookup("table2")
    # Turn on validation checking
    tables.enable_checks(True)
    table3 = tables.add("table3")
    assert table3._checking_enabled is True
    # Clear the stored symbol tables
    tables.clear()
    assert tables._current_scope is None
    assert tables._symbol_tables == {}


def test_str_method():
    """Tests for the str() method."""
    tables = SymbolTables()
    text = str(tables)
    assert "SymbolTables: 0 tables" in text
    tables.enter_scope("some_mod")
    tables.exit_scope()
    tables.enter_scope("other_mod")
    text = str(tables)
    assert (
        "SymbolTables: 2 tables\n"
        "========================\n"
        "other_mod\nsome_mod" in text
    )
    tables.clear()
    assert "SymbolTables: 0 tables" in str(tables)


def test_nested_scoping():
    """Test the functionality related to moving into and out of scoping
    regions."""
    tables = SymbolTables()
    tables.enter_scope("some_mod")
    outer_table = tables.lookup("some_mod")
    assert tables.current_scope is outer_table
    tables.exit_scope()
    assert tables.current_scope is None
    # Enter scope of existing symbol table
    tables.enter_scope("some_mod")
    assert tables.current_scope is outer_table
    tables.enter_scope("some_func")
    assert tables.current_scope.parent is outer_table
    tables.exit_scope()
    assert tables.current_scope is outer_table
    tables.enter_scope("some_func")
    # The new symbol table should not appear in the top-level symbol tables
    # because it is nested.
    with pytest.raises(KeyError) as err:
        tables.lookup("some_func")
    assert "some_func" in str(err.value)
    assert tables.current_scope.parent is outer_table
    tables.exit_scope()
    tables.exit_scope()
    assert tables.current_scope is None
    # Attempting to call exit_scope again should cause an error
    with pytest.raises(SymbolTableError) as err:
        tables.exit_scope()
    assert "exit_scope() called but no current scope exists" in str(err.value)


def test_nested_removal():
    """Tests the removal of symbol tables when we have nested scopes."""
    tables = SymbolTables()
    tables.enter_scope("some_mod")
    tables.enter_scope("some_func")
    # Cannot find a symbol table if we are currently inside it
    with pytest.raises(SymbolTableError) as err:
        tables.remove("some_func")
    assert (
        "Failed to find a table named 'some_func' in either the current "
        "scope (which contains []) or the list of top-level symbol tables "
        "(['some_mod'])." in str(err.value)
    )
    tables.exit_scope()
    tables.remove("some_func")
    assert tables.current_scope.children == []
    tables.enter_scope("another_func")
    # We should not be able to remove a symbol table that is a parent of the
    # current scope.
    with pytest.raises(SymbolTableError) as err:
        tables.remove("some_mod")
    assert (
        "Cannot remove top-level symbol table 'some_mod' because the "
        "current scope 'another_func' has it as an ancestor" in str(err.value)
    )
    # Create a 2nd, top-level symbol table.
    tables.exit_scope()
    tables.exit_scope()
    assert tables.current_scope is None
    tables.enter_scope("another_mod")
    # Should be able to remove the other, top-level symbol table
    tables.remove("some_mod")
    assert "some_mod" not in tables._symbol_tables
    assert "another_mod" in tables._symbol_tables



# Copyright (c) 2023 Science and Technology Facilities Council.

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R552 : This file tests the support for the
NAMELIST statement.

"""

import pytest

from fparser.two.Fortran2003 import Namelist_Stmt
from fparser.two.utils import NoMatchError


@pytest.mark.parametrize(
    "code",
    [
        "",
        "namelost",
        "namelist",
        "namelist x",
        "namelist x/",
        "namelist /x",
        "namelist /x/ y /",
    ],
)
def test_match_errors(code):
    """Test that the match method returns None when the supplied code is
    invalid. Also check that this results in a NoMatchError for an
    instance of the class.

    """
    assert not Namelist_Stmt.match(code)
    with pytest.raises(NoMatchError):
        _ = Namelist_Stmt(code)


def test_simple():
    """Test that a namelist with a single name and list is matched and
    that the tostr() method outputs the resultant code as
    expected. Also check that the namelist keyword matching is case
    insensitive and that leading and trailing spaces are supported.

    """
    result = Namelist_Stmt(" NamelisT /x/ a ")
    assert isinstance(result, Namelist_Stmt)
    assert result.tostr() == "NAMELIST /x/ a"


def test_multi():
    """Test that multiple names and lists are matched, with and without a
    comma separator and that the tostr() method outputs the resultant
    code as expected.

    """
    result = Namelist_Stmt("namelist /x/ a, /y/ b,c /z/ d")
    assert isinstance(result, Namelist_Stmt)
    assert result.tostr() == "NAMELIST /x/ a, /y/ b, c, /z/ d"



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Open_Stmt rule R904
"""

from fparser.two.Fortran2003 import Open_Stmt as Open_Stmt_2003
from fparser.two.utils import CALLBase


class Open_Stmt(Open_Stmt_2003):  # R904
    """
    Fortran2008 Rule R904.

    open-stmt is OPEN ( connect-spec-list )

    """

    subclass_names = []
    use_names = ["Connect_Spec_List"]

    @staticmethod
    def match(string):
        """
        Attempts to match the supplied string as an Open_Stmt.

        :param str string: the string to attempt to match.

        :returns: a new Open_Stmt object if the match is successful, None otherwise.
        :rtype: Optional[:py:class:`fparser.two.Fortran2008.Open_Stmt]

        """
        # Avoid circular dependencies by importing here.
        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Connect_Spec_List

        obj = CALLBase.match("OPEN", Connect_Spec_List, string, require_rhs=True)
        if not obj:
            return None

        # Apply constraints now that we have the full Connect_Spec_List.
        have_unit = False
        have_newunit = False
        connect_specs = []
        spec_list = obj[1].children
        for spec in spec_list:
            if spec.children[0] in connect_specs:
                # C903 - no specifier can appear more than once.
                return None
            connect_specs.append(spec.children[0])
            if spec.children[0] == "UNIT":
                have_unit = True
            elif spec.children[0] == "NEWUNIT":
                have_newunit = True
            if have_unit and have_newunit:
                # C906 - cannot have both UNIT and NEWUNIT
                return None
        if not (have_unit or have_newunit):
            # C904 - a file unit number must be specified.
            return None
        return obj



# Copyright (c) 2023 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R1206

procedure-stmt is [ MODULE ] PROCEDURE [ :: ] procedure-name-list

"""

import pytest
from fparser.two.Fortran2008 import Procedure_Stmt
from fparser.two.utils import NoMatchError


def test_start_space():
    """Test that there is a match if the string contains white space at
    the start and that the tostr() output is as expected.

    """
    result = Procedure_Stmt(" procedure dummy")
    assert isinstance(result, Procedure_Stmt)
    assert str(result) == "PROCEDURE dummy"


def test_module():
    """Test that there is a match if the string contains the optional
    MODULE keyword and that the tostr() output is as expected.

    """
    result = Procedure_Stmt(" module procedure dummy")
    assert isinstance(result, Procedure_Stmt)
    assert str(result) == "MODULE PROCEDURE dummy"


def test_colons():
    """Test that there is a match if the string contains optional :: after
    the procedure keyword and that the tostr() output is as expected.

    """
    result = Procedure_Stmt(" module procedure :: dummy")
    assert isinstance(result, Procedure_Stmt)
    assert str(result) == "MODULE PROCEDURE :: dummy"


@pytest.mark.parametrize(
    "string",
    [
        "procedur dummy",
        "modul procedure dummy",
        "procedure : dummy",
        "procedure ",
        "procedure :: ",
    ],
)
def test_invalid(string):
    """Test that there is no match for various invalid input strings."""
    with pytest.raises(NoMatchError):
        _ = Procedure_Stmt(string)



# Copyright (c) 2020 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2008 rule R437.

component-attr-spec = access-spec
                      or ALLOCATABLE
                      or CODIMENSION lbracket coarray-spec rbracket
                      or CONTIGUOUS
                      or DIMENSION ( component-array-spec )
                      or POINTER
"""

import pytest
from fparser.two.Fortran2003 import (
    Access_Spec,
    Dimension_Component_Attr_Spec,
    NoMatchError,
)
from fparser.two.Fortran2008 import Component_Attr_Spec, Codimension_Attr_Spec
from fparser.api import get_reader


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize("mixed_case_ref", ["ALlOCaTABLE", "CONTIGUOUs", "PoInTeR"])
def test_component_attr_spec(mixed_case_ref):
    """Test the parsing of component attributes provided as mixed case, upper
    case and lower case strings."""
    for manip in [str, str.upper, str.lower]:
        obj = Component_Attr_Spec(manip(mixed_case_ref))
        assert isinstance(obj, Component_Attr_Spec), repr(obj)
        assert str(obj) == mixed_case_ref.upper()


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr",
    ["", " ALLOCATABLE", "POINTER ", "POIN TER", "CONTIGOUS", "ALOCATABLE", "FOO"],
)
def test_invalid_component_attr_spec(attr):
    """Test that invalid or misspelled component attributes raise exception."""
    with pytest.raises(NoMatchError):
        _ = Component_Attr_Spec(attr)


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr, ref",
    [
        ("DIMENSION   (:)", "DIMENSION(:)"),
        ("dimenSION(  : )", "DIMENSION(:)"),
        ("DIMENSION (5)", "DIMENSION(5)"),
        ("dimension (1:5, 2, 3:4)", "DIMENSION(1 : 5, 2, 3 : 4)"),
    ],
)
def test_component_attr_spec_dimension(attr, ref):
    """Test the parsing of dimension attribute in component_attr_spec."""
    obj = Component_Attr_Spec(attr)
    assert isinstance(obj, Dimension_Component_Attr_Spec), repr(obj)
    assert str(obj) == ref


@pytest.mark.usefixtures("f2008_create")
@pytest.mark.parametrize(
    "attr, ref",
    [
        ("codimension[*]", "CODIMENSION [*]"),
        ("CODIMENSION   [:]", "CODIMENSION [:]"),
        ("COdimenSION[  : ]", "CODIMENSION [:]"),
        ("coDIMENSIon [ *]", "CODIMENSION [*]"),
        (" CODIMENSION [1:5, *]   ", "CODIMENSION [1 : 5, *]"),
        ("codimension [1:5, 2, 3:*]", "CODIMENSION [1 : 5, 2, 3 : *]"),
    ],
)
def test_component_attr_spec_codimensions(attr, ref):
    """Test the parsing of codimension attribute in component_attr_spec."""
    obj = Component_Attr_Spec(attr)
    assert isinstance(obj, Codimension_Attr_Spec), repr(obj)
    assert str(obj) == ref


@pytest.mark.usefixtures("f2008_create")
def test_private_attr():
    """Tests the parsing of private attribute."""
    obj = Component_Attr_Spec("private")
    assert isinstance(obj, Access_Spec), repr(obj)
    assert str(obj) == "PRIVATE"


def test_component_attr_spec_list_parser(f2008_parser):
    """Test that Component_Attr_Spec_List is generated correctly
    and used by the F2008 parser."""
    code = """
MODULE FOO
  TYPE :: GRID_TYPE
    REAL, ALLOCATABLE, CODIMENSION [:, :, :] :: GRID(:, :, :)
    REAL :: BAR
  END TYPE GRID_TYPE
END MODULE FOO
    """.strip()
    reader = get_reader(code)
    result = f2008_parser(reader)
    assert str(result) == code



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Action_Stmt rule R214
"""

from fparser.two.Fortran2003 import Action_Stmt as Action_Stmt_2003


class Action_Stmt(Action_Stmt_2003):  # R214
    """
    Fortran 2008 rule R214.

    .. code-block:: fortran

        action-stmt is allocate-stmt
                        or assignment-stmt
                        or backspace-stmt
                        or call-stmt
                        or close-stmt
                        or continue-stmt
                        or cycle-stmt
                        or deallocate-stmt
                        or end-function-stmt
                        or end-mp-subprogram-stmt
                        or end-program-stmt
                        or end-subroutine-stmt
                        or endfile-stmt
                        or error-stop-stmt
                        or exit-stmt
                        or flush-stmt
                        or forall-stmt
                        or goto-stmt
                        or if-stmt
                        or inquire-stmt
                        or lock-stmt
                        or nullify-stmt
                        or open-stmt
                        or pointer-assignment-stmt
                        or print-stmt
                        or read-stmt
                        or return-stmt
                        or rewind-stmt
                        or stop-stmt
                        or sync-all-stmt
                        or sync-images-stmt
                        or sync-memory-stmt
                        or unlock-stmt
                        or wait-stmt
                        or where-stmt
                        or write-stmt
                        or arithmetic-if-stmt
                        or computed-goto-stmt

    The implementation of this rule adds the relevant subclass names
    for new statements added in Fortran 2008.

    Associated constraints are:

    "C201 (R208) An execution-part shall not contain an end-function-stmt,
          end-mp-subprogram-stmt, end-program-stmt, or end-subroutine-stmt."

    NB: The following statements are not yet implemented:
    end-mp-subprogram-stmt, endfile-stmt, lock-stmt, sync-all-stmt,
    sync-images-stmt, sync-memory-stmt, unlock-stmt.

    """

    # Fortran 2008 adds a few additional action-stmt. We therefore
    # extend the Fortran 2003 specification
    subclass_names = Action_Stmt_2003.subclass_names[:]
    subclass_names.append("Error_Stop_Stmt")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import subprocess
import fparser

project = 'fparser'
copyright = '2017-2024, Science and Technology Facilities Council'
author = 'Andrew Porter, Rupert Ford, Balthasar Reuter, Joerg Henrichs and Pearu Peterson'

version = fparser._get_version()
release = fparser._get_version()

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.viewcode',
    'autoapi.extension',
]

autoapi_dirs = ['../../src']
autoapi_ignore = ['*test_*']
autoapi_keep_files = True

templates_path = ['_templates']
exclude_patterns = []

root_doc = 'index'

language = 'en'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
# We specify the directory containing the doxygen output (as configured in
# ../doxygen.config) so that we can link to it from index.rst.
html_static_path = ['doxygen']

# -- Options for intersphinx extension ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
}

# Generate the Doxygen documentation
subprocess.call('cd ..; doxygen doxygen.config', shell=True)


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
#  documentclass [howto/manual]).
latex_documents = [
    ('index', 'fparser.tex', 'fparser Documentation',
     'Andrew Porter, Rupert Ford, Balthasar Reuter, \\\\ '
     'Joerg Henrichs and Pearu Peterson', 'manual'),
]



# Modified work Copyright (c) 2017-2018 Science and Technology
# Facilities Council
# Original work Copyright (c) 1999-2008 Pearu Peterson
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

"""
Test fparser support for parsing select-case and select-type blocks

"""

import pytest
import fparser.common.sourceinfo


def test_case():
    """Basic tests for parsing of individual case statements"""
    from fparser.one.tests.test_parser import parse
    from fparser.one.block_statements import Case

    assert parse(Case, "case (1)") == "CASE ( 1 )"
    assert parse(Case, "case (1:)") == "CASE ( 1 : )"
    assert parse(Case, "case (:1)") == "CASE ( : 1 )"
    assert parse(Case, "case (1:2)") == "CASE ( 1 : 2 )"
    assert parse(Case, "case (a(1,2))") == "CASE ( a(1,2) )"
    assert parse(Case, 'case ("ab")') == 'CASE ( "ab" )'
    assert parse(Case, "case default") == "CASE DEFAULT"
    assert parse(Case, "case (1:2 ,3:4)") == "CASE ( 1 : 2, 3 : 4 )"
    assert parse(Case, "case (a(1,:):)") == "CASE ( a(1,:) : )"
    assert parse(Case, "case default") == "CASE DEFAULT"


def test_case_internal_error(monkeypatch, capsys):
    """Check that expected errors are raised when invalid case
    statements are encountered"""
    from fparser.one.block_statements import Case
    from fparser.common.readfortran import FortranStringReader

    reader = FortranStringReader("CASE (yes)")
    reader.set_format(fparser.common.sourceinfo.FortranFormat(True, False))
    item = next(reader)
    stmt = Case(item, item)
    # Monkeypatch our valid Case object so that get_line() now
    # returns something invalid. We have to do it this way
    # because if we started with this text then we wouldn't get
    # past the match() method
    monkeypatch.setattr(stmt.item, "get_line", lambda: "case invalid")
    # Monkeypatch the Case object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to print
    monkeypatch.setattr(stmt, "warning", lambda out: print(out))
    stmt.process_item()
    output, _ = capsys.readouterr()
    print(output)
    assert "Internal error when parsing CASE statement" in output


def test_class_internal_error(monkeypatch, capsys):
    """Check that expected errors are raised when invalid CLASS
    statements are encountered"""
    from fparser.one.block_statements import ClassIs
    from fparser.common.readfortran import FortranStringReader

    reader = FortranStringReader("CLASS IS (yes)")
    reader.set_format(fparser.common.sourceinfo.FortranFormat(True, False))
    item = next(reader)
    stmt = ClassIs(item, item)
    # Monkeypatch our valid Case object so that get_line() now
    # returns something invalid. We have to do it this way
    # because if we started with this text then we wouldn't get
    # past the match() method
    monkeypatch.setattr(stmt.item, "get_line", lambda: "class invalid")
    # Monkeypatch the Case object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to with a call to print
    monkeypatch.setattr(stmt, "warning", lambda out: print(out))
    stmt.process_item()
    output, _ = capsys.readouterr()
    assert "Internal error when parsing CLASS statement" in output


def test_select_case():
    """Test that fparser correctly recognises select case"""
    from fparser import api

    source_str = """
    subroutine foo
    integer :: iflag = 1
    real    :: aval = 0.0
    select case(iflag)
    case(1)
      aval = 1.0
    case default
      aval = 0.0
    end select
    ! Same again but with more white space
    select   case(iflag)
    case ( 1 )
      aval = 1.0
    case  default
      aval = 0.0
    end select
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False, ignore_comments=True)
    assert tree
    select_list = []
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectCase):
            select_list.append(statement)
    assert len(select_list) == 2
    for statement in select_list:
        assert isinstance(statement, fparser.one.block_statements.SelectCase)
        assert isinstance(statement.content[0], fparser.one.statements.Case)
        assert isinstance(statement.content[2], fparser.one.statements.Case)
        assert isinstance(statement.content[3], fparser.one.statements.Assignment)
    gen = str(tree)
    print(gen)
    assert "SELECT CASE ( iflag )" in gen


@pytest.mark.xfail(reason="fparser does not work with named select statements")
def test_named_select_case():
    """Test that fparser correctly recognises a named select case"""
    from fparser import api

    source_str = """
    subroutine foo
    integer :: iflag = 1
    real    :: aval = 0.0
    incase: select case(iflag)
    case(1) incase
      aval = 1.0
    case default   incase
      aval = 0.0
    end select  incase
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    select_list = []
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectCase):
            select_list.append(statement)
    assert len(select_list) == 2
    for statement in select_list:
        assert isinstance(statement, fparser.one.block_statements.SelectCase)
        assert isinstance(statement.content[0], fparser.one.statements.Case)
        assert isinstance(statement.content[2], fparser.one.statements.Case)
        assert isinstance(statement.content[3], fparser.one.statements.Assignment)
    gen = str(tree)
    print(gen)
    assert "incase: SELECT CASE ( iflag )" in gen


def test_select_case_brackets():
    """Test that fparser correctly parses a select case involving
    parentheses"""
    from fparser import api

    source_str = """
    subroutine foo
    integer :: iflag(2) = 1
    real    :: aval = 0.0
    select case(iflag(1))
    case(1)
      aval = 1.0
    case default
      aval = 0.0
    end select
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keep pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectCase):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectCase)
    assert isinstance(statement.content[0], fparser.one.statements.Case)
    assert isinstance(statement.content[2], fparser.one.statements.Case)
    assert isinstance(statement.content[3], fparser.one.statements.Assignment)
    gen = str(tree)
    print(gen)
    assert "SELECT CASE ( iflag(1) )" in gen


def test_select_type():
    """Test that fparser correctly recognises select type"""
    from fparser import api

    source_str = """
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    type is (some_type)
      aval = 1.0
    class is (some_class)
      aval = 0.0
    class default
      aval = -1.0
    end select
    ! Same again but with more white space
    select   type ( an_object )
    type  is  ( some_type(i), some_other_type )
      aval = 1.0
    class  is  ( some_class(1), some_other_class )
      aval = 0.0
    class   default
      aval = -1.0
    end select
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    select_list = []
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            select_list.append(statement)
    assert len(select_list) == 2
    for statement in select_list:
        assert isinstance(statement, fparser.one.block_statements.SelectType)
        assert isinstance(statement.content[0], fparser.one.statements.TypeIs)
        assert isinstance(statement.content[2], fparser.one.statements.ClassIs)
        assert isinstance(statement.content[3], fparser.one.statements.Assignment)
        assert isinstance(statement.content[4], fparser.one.statements.ClassIs)
    gen = str(tree)
    print(gen)
    assert "SELECT TYPE ( an_object )" in gen
    assert "TYPE IS ( some_type )" in gen
    assert "TYPE IS ( some_type(i), some_other_type )" in gen
    assert "CLASS IS ( some_class )" in gen
    assert "CLASS DEFAULT" in gen
    assert "CLASS IS ( some_class(1), some_other_class )" in gen


def test_type_is_process_item(monkeypatch, capsys):
    """Test error condition raised in TypeIs.process_item() method"""
    from fparser import api

    source_str = """
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    type is (some_type)
      aval = 1.0
    class is (some_class)
      aval = 0.0
    end select
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.TypeIs)
    typeis = statement.content[0]
    typeis.parent.name = "not_a_name"
    monkeypatch.setattr(typeis.item, "get_line", lambda: "type is (blah): wrong_name")
    # Monkeypatch the typeis object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to with a call to print
    monkeypatch.setattr(typeis, "warning", lambda out: print(out))
    typeis.process_item()
    output, _ = capsys.readouterr()
    print(output)
    assert "expected type-is-construct-name 'not_a_name' but got " in output


def test_type_is_to_fortran():
    """Test error condition raised in TypeIs.to_fortran() method"""
    from fparser import api
    from fparser.common.utils import ParseError

    source_str = """
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    type is (some_type)
      aval = 1.0
    class is (some_class)
      aval = 0.0
    end select
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.TypeIs)
    typeis = statement.content[0]
    typeis.name = "some_name"
    fort = typeis.tofortran()
    assert "TYPE IS ( some_type ) some_name" in fort
    # Now break the internal state and check for expected exception
    typeis.items = None
    with pytest.raises(ParseError) as excinfo:
        _ = typeis.tofortran()
    assert "TYPE IS construct must have arguments" in str(excinfo.value)


def test_class_is_process_item(monkeypatch, capsys):
    """Test error condition raised in ClassIs.process_item() method"""
    from fparser import api

    source_str = """
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    class is (some_class)
      aval = 0.0
    class default
      aval = -1.0
    end select
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.ClassIs)
    clsis = statement.content[0]
    clsis.parent.name = "not_a_name"
    monkeypatch.setattr(clsis.item, "get_line", lambda: "class is (blah): wrong_name")
    # Monkeypatch the typeis object so that a call to self.warning
    # (which normally results in a call to the logger) gets replaced
    # with a call to with a call to print
    monkeypatch.setattr(clsis, "warning", lambda out: print(out))
    clsis.process_item()
    output, _ = capsys.readouterr()
    print(output)
    assert "expected class-construct-name 'not_a_name' but got " in output


def test_class_is_to_fortran():
    """Test ClassIs.to_fortran() method"""
    from fparser import api

    source_str = """
    subroutine foo(an_object)
    class(*) :: an_object
    real    :: aval = 0.0
    select type(an_object)
    class is (some_class)
      aval = 0.0
    end select
    end subroutine foo
    """
    tree = api.parse(source_str, isfree=True, isstrict=False)
    assert tree
    statement = None  # Keeps pylint happy
    for statement in tree.content[0].content:
        if isinstance(statement, fparser.one.block_statements.SelectType):
            break
    assert isinstance(statement, fparser.one.block_statements.SelectType)
    assert isinstance(statement.content[0], fparser.one.statements.ClassIs)
    clsis = statement.content[0]
    clsis.name = "some_name"
    fort = clsis.tofortran()
    assert "CLASS IS ( some_class ) some_name" in fort



# Copyright (c) 2017-2023 Science and Technology Facilities Council.
# Original work Copyright (c) 1999-2008 Pearu Peterson

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# --------------------------------------------------------------------

# The original software (in the f2py project) was distributed under
# the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

#   a. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   b. Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#   c. Neither the name of the F2PY project nor the names of its
#      contributors may be used to endorse or promote products derived from
#      this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

# First version by: Pearu Peterson <pearu@cens.ioc.ee>
# First created: Oct 2006

from importlib.metadata import PackageNotFoundError
from importlib import metadata

import logging
import codecs


def _get_version():
    """
    :returns: the version of this package.
    :rtype: str
    """
    try:
        return metadata.version(__name__)
    except PackageNotFoundError:
        # Package is not installed.
        from setuptools_scm import get_version

        return get_version(root="../..", relative_to=__file__)


__version__ = _get_version()

logging.getLogger(__name__).addHandler(logging.NullHandler())


def log_decode_error_handler(err):
    """
    A custom error handler for use when reading files. Removes any
    characters that cause decoding errors and logs the error.

    :returns: 2-tuple containing replacement for bad chars (an empty string \
              and the position from where encoding should continue.
    :rtype: Tuple[str, int]

    """
    message = f"character in input file. Error returned was {str(err)}."
    # Log the fact that this character will be removed from the input file
    logging.getLogger(__name__).warning("Skipped bad %s", message)
    return ("", err.end)


codecs.register_error("fparser-logging", log_decode_error_handler)



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Type_Declaration_Stmt rule R501
"""

from fparser.two.Fortran2003 import Type_Declaration_Stmt as Type_Declaration_Stmt_2003


class Type_Declaration_Stmt(Type_Declaration_Stmt_2003):  # R501
    """
    Fortran 2008 rule 501.

    .. code-block:: fortran

        type-declaration-stmt is declaration-type-spec [ [ , attr-spec ] ... :: ]
                                 entity-decl-list

    The implementation of this rule does not add anything to the Fortran 2003
    variant but overwrites :py:meth:`get_attr_spec_list_cls` to use
    the Fortran 2008 variant of :py:class:`Attr_Spec_List`.

    Associated constraints are:

    "C501 (R501)  The same attr-spec shall not appear more than once in a given
          type-declaration-stmt."
    "C502 (R501)  If a language-binding-spec with a NAME= specifier appears,
          the entity-decl-list shall consist of a single entity-decl."
    "C503 (R501)  If a language-binding-spec is specified, the entity-decl-list
          shall not contain any procedure names."
    "C505 (R501)  If initialization appears, a double-colon separator shall
          appear before the entity-decl-list."

    C501-C503, C505 are currently not checked - issue #259.

    """

    @staticmethod
    def get_attr_spec_list_cls():
        """Return the type used to match the attr-spec-list

        This overwrites the Fortran 2003 type with the Fortran 2008 variant.

        """
        # Avoid circular dependencies by importing here.
        # pylint: disable=import-outside-toplevel
        from fparser.two.Fortran2008 import Attr_Spec_List

        return Attr_Spec_List



# -----------------------------------------------------------------------------
# BSD 3-Clause License
#
# Copyright (c) 2023, Science and Technology Facilities Council.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

"""
Module containing Fortran2008 Block_Stmt rule R808
"""

from fparser.two.utils import StmtBase, WORDClsBase, ScopingRegionMixin


class Block_Stmt(StmtBase, WORDClsBase, ScopingRegionMixin):
    """
    Fortran 2008 Rule 808.

    block-stmt is [ block-construct-name : ] BLOCK

    """

    subclass_names = []
    use_names = ["Block_Construct_Name"]
    counter = 0

    @staticmethod
    def match(string):
        """
        Attempts to match the supplied text with this rule.

        :param str string: the text to match.

        :returns: a tuple of the matched node and instance of Counter or \
                  None if there is no match.
        :rtype: Tuple["BLOCK", \
                      :py:class:`fparser.two.Fortran2008.Block_Stmt.Counter`] \
                | NoneType
        """
        found = WORDClsBase.match("BLOCK", None, string)
        if not found:
            return None
        block, _ = found
        # Construct a unique name for this BLOCK (in case it isn't named). We
        # ensure the name is not a valid Fortran name so that it can't clash
        # with any regions named in the code.
        scope_name = f"block:{Block_Stmt.counter}"
        Block_Stmt.counter += 1
        # TODO #397. Ideally we'd have the name associated with the Block
        # Construct here (if any) so that it could be displayed in repr.
        # As it is, repr will show scope_name which will not be the same
        # as any explicit name given to the Block. (This name *is* shown
        # in the repr of the End_Block_Stmt.) This problem is common to
        # other block constructs such as Block_Nonlabel_Do_Construct.
        return block, scope_name

    def get_scope_name(self):
        """
        :returns: the name of this scoping region.
        :rtype: str
        """
        if self.item.name:
            return self.item.name
        return self.items[1]

    def get_start_name(self):
        """
        :returns: the name associated with this Block construct or None.
        :rtype: str | NoneType
        """
        return self.item.name

    def tostr(self):
        """
        :returns: the string representation of this node.
        :rtype: str
        """
        return "BLOCK"



# Copyright (c) 2018-2021 Science and Technology Facilities Council

# All rights reserved.

# Modifications made as part of the fparser project are distributed
# under the following license:

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test Fortran 2003 rule R1102 : This file tests the support for the
program statement.

"""

import pytest
from fparser.api import get_reader
from fparser.two.utils import NoMatchError
from fparser.two.symbol_table import SYMBOL_TABLES
from fparser.two.Fortran2003 import Program_Stmt, Program, Name


@pytest.mark.usefixtures("f2003_create")
def test_valid():
    """Test that valid code is parsed correctly."""

    obj = Program_Stmt("program a")
    assert isinstance(obj, Program_Stmt)
    assert str(obj) == "PROGRAM a"
    assert repr(obj) == "Program_Stmt('PROGRAM', Name('a'))"
    # Check that the parent of the Name is correctly set
    assert obj.items[1].parent is obj


@pytest.mark.usefixtures("f2003_create")
def test_invalid():
    """Test that exceptions are raised for invalid code."""

    for string in [
        "",
        "  ",
        "prog",
        "program",
        "programa",
        "a program",
        "a program a",
        "program a a",
    ]:
        with pytest.raises(NoMatchError) as excinfo:
            _ = Program_Stmt(string)
        assert "Program_Stmt: '{0}'".format(string) in str(excinfo.value)


@pytest.mark.usefixtures("f2003_create")
def test_prog_symbol_table():
    """Check that an associated symbol table is created when parsing a
    program unit."""
    reader = get_reader("program my_prog\n" "end program my_prog\n")
    prog = Program(reader)
    assert "my_prog" in SYMBOL_TABLES._symbol_tables


def test_get_name():
    """Test we can get the name of the program"""
    obj = Program_Stmt("program foo")
    assert obj.get_name() == Name("foo")


def test_get_start_name():
    """Test we can get the name of the function as a string"""

    obj = Program_Stmt("program foo")
    assert obj.get_start_name() == "foo"



# Modified work Copyright (c) 2022 Science and Technology
# Facilities Council.
#
# All rights reserved.
#
# Modifications made as part of the fparser project are distributed
# under the following license:
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import pytest
from fparser.api import get_reader
from fparser.two.Fortran2003 import Select_Type_Construct
from fparser.two.utils import FortranSyntaxError


def test_select_type_construct():
    """Check 'select type' construct is parsed correctly"""
    tcls = Select_Type_Construct
    tree = tcls(
        get_reader(
            """\
            n:SELECT TYPE ( A => P_OR_C )
            CLASS IS ( POINT ) n
            PRINT *, A%X, A%Y ! This block gets executed
            TYPE IS ( POINT_3D ) n
            PRINT *, A%X, A%Y, A%Z
            END SELECT n
            """,
            ignore_comments=False,
        )
    )
    assert (
        str(tree) == "n:SELECT TYPE(A=>P_OR_C)\n"
        "  CLASS IS (POINT) n\n"
        "  PRINT *, A % X, A % Y\n"
        "  ! This block gets executed\n"
        "  TYPE IS (POINT_3D) n\n"
        "  PRINT *, A % X, A % Y, A % Z\n"
        "END SELECT n"
    )


def test_select_type_construct_wrong_name(f2003_create, fake_symbol_table):
    """Check named 'select type' construct has correct start/end names"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Select_Type_Construct(get_reader("""\
            name: select type (n)
            type is (point)
                a = 1
            end select wrong"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")


def test_select_type_construct_missing_start_name(f2003_create, fake_symbol_table):
    """Check named 'select type' construct has correct start/end names"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Select_Type_Construct(get_reader("""\
            select type(n)
            type is (point)
                a = 1
            end select name"""))
    assert exc_info.value.args[0].endswith(
        "Name 'name' has no corresponding starting name"
    )


def test_select_type_construct_missing_end_name(f2003_create, fake_symbol_table):
    """Check named 'select type' construct has correct start/end names"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Select_Type_Construct(get_reader("""\
            name: select type(n)
            type is (point)
                a = 1
            end select"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name' but none given")


def test_select_type_construct_select_type_wrong_name(f2003_create, fake_symbol_table):
    """Check named 'select type' construct has correct start/end names"""
    with pytest.raises(FortranSyntaxError) as exc_info:
        Select_Type_Construct(get_reader("""\
            name: select type(n)
            type is (point) wrong
                a = 1
            end select name"""))
    assert exc_info.value.args[0].endswith("Expecting name 'name', got 'wrong'")
